// ChildView.cpp : implementation of the CChildView class

#include "stdafx.h"
#include "MJDict.h"
#include "ChildView.h"
#include "DictOptions.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

extern CMJDictApp theApp;

/////////////////////////////////////////////////////////////////////////////
// CChildView

CChildView::CChildView()
{
	lastsearch= _T("");
	m_Input = _T("");
	m_Output = _T("");
}

CChildView::~CChildView()
{
	// Finished with the font, now delete the font object.
	font.DeleteObject();
}

BEGIN_MESSAGE_MAP(CChildView,CWnd )
	//{{AFX_MSG_MAP(CChildView)
	ON_WM_PAINT()
	ON_COMMAND(ID_VERB, OnVerb)
	ON_COMMAND(ID_KANJI, OnKanji)
	ON_COMMAND(ID_MAINDICT, OnMaindict)
	ON_COMMAND(ID_ALTDICT, OnAltdict)
	ON_COMMAND(ID_EDIT_COPY, OnEditCopy)
	ON_COMMAND(ID_EDIT_CUT, OnEditCut)
	ON_COMMAND(ID_EDIT_PASTE, OnEditPaste)
	ON_WM_SIZE()
	ON_COMMAND(ID_SEARCHRESULTS, OnSearchresults)
	ON_COMMAND(ID_DICTIONARIES, OnDictionaries)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CChildView message handlers

BOOL CChildView::PreCreateWindow(CREATESTRUCT& cs) 
{
	if (!CWnd::PreCreateWindow(cs))
		return FALSE;

#ifndef UNDER_CE
	cs.dwExStyle |= WS_EX_CLIENTEDGE;
#endif
	cs.style &= ~WS_BORDER;
	cs.lpszClass = AfxRegisterWndClass(CS_HREDRAW|CS_VREDRAW|CS_DBLCLKS, 
#ifdef UNDER_CE
		NULL,
#else
		::LoadCursor(NULL, IDC_ARROW),
#endif
		HBRUSH(COLOR_WINDOW+1), NULL);

	return TRUE;
}

BOOL CChildView::Create(LPCTSTR lpszClassName, LPCTSTR lpszWindowName, DWORD dwStyle, const RECT& rect, CWnd* pParentWnd, UINT nID, CCreateContext* pContext) 
{
	// TODO: Add your specialized code here and/or call the base class
	
	BOOL result = CWnd::Create(lpszClassName, lpszWindowName, dwStyle, rect, pParentWnd, nID, pContext);

	InputArea.Create(WS_CHILD|WS_VISIBLE|WS_BORDER|WS_VSCROLL|CBS_DROPDOWN|CBS_AUTOHSCROLL, 
		CRect(0, 0, 240, 100), this, IDC_INPUT);
	InputArea.InitStorage(50,50);
	RECT iRect;
	InputArea.GetClientRect(&iRect);
	OutputArea.Create(WS_CHILD|WS_VISIBLE|WS_BORDER|ES_MULTILINE|ES_READONLY|WS_VSCROLL|ES_AUTOVSCROLL, 
		CRect(0, iRect.bottom, 240, 188), this, IDC_OUTPUT);

	InputArea.SetFocus(); // Set focus to input area

	// Initializes a CFont object with the characteristics given 
	// in a LOGFONT structure.
	LOGFONT lf;
	memset(&lf, 0, sizeof(LOGFONT));       // Zero out the structure.
	lf.lfHeight = -13;
	lstrcpy(lf.lfFaceName, _T("MS Gothic"));
	lf.lfCharSet = SHIFTJIS_CHARSET;
	font.CreateFontIndirect(&lf);
	InputArea.SetFont(&font);
	OutputArea.SetFont(&font);

	return result;
}

void CChildView::OnPaint() 
{
	CPaintDC dc(this); // device context for painting
	
	// TODO: Add your message handler code here
	
	// Do not call CWnd::OnPaint() for painting messages
}

void CChildView::OnMaindict() 
{
	CWaitCursor wait;

	// Get file path and name from the registry and then open the dictionaries
	CString dict, index;
    dict = theApp.GetProfileString(SECTION, _T("EDict"), _T("\\Program Files\\edict.euc"));
    index = theApp.GetProfileString(SECTION, _T("EDictIndex"), _T("\\Program Files\\edict.jdx"));
	CDict edict;
	if (!initSearch(edict.InitDict(dict, index)))
		return;

	// Check type of input string
	inputType=2; // kana
	if (m_Input.GetAt(0)>=0x4000) { // Kanji
		inputType=0;
	}
	if (m_Input.GetAt(0)<=0x0100) { // Roman
		inputType=2;
	}

	int c = 0;
	CString result = edict.Find(inputType+1, m_Input);
	while (c<600 && result.GetLength()!=0) {
		c++;
		m_Output += result;
		if (c<12 || c%20==0)
			OutputArea.SetWindowText(m_Output);
		result = edict.Find(inputType, m_Input);
	}

	if (m_Output.GetLength()==0) {
		m_Output = _T("Nothing found.");
	}

	OutputArea.SetWindowText(m_Output);
	InputArea.SetEditSel(0,-1);
	InputArea.SetFocus();
}

bool CChildView::initSearch(bool r)
{
	// check result from loading dictionary
	if (!r) {
		m_Output = _T("Dictionary not found");
		OutputArea.SetWindowText(m_Output);
		InputArea.SetEditSel(0,-1);
		InputArea.SetFocus();
		return false;
	}

	// If OutputArea has focus, use the selected text as the input
	if (GetFocus() == (CWnd*)&OutputArea) {
		OutputArea.Copy();
		InputArea.SetEditSel(0,-1);
		InputArea.Paste();
	}

	// Get input string and check its not empty
	InputArea.GetWindowText(m_Input);
	if (m_Input.GetLength()==0) {
		OutputArea.SetWindowText(_T("Enter characters to search for."));
		InputArea.SetFocus();
		return false;
	}

	// Save search history
	if (m_Input!=lastsearch) {
		lastsearch=m_Input;
		InputArea.InsertString(0,m_Input);
	}

	// Display wait message
	OutputArea.SetWindowText(_T("Wait..."));
	m_Output = _T("");
	return true;
}

void CChildView::OnKanji() 
{
	CWaitCursor wait;
	// Get file path and name from the registry and then open the dictionaries
	CString dict;
    dict = theApp.GetProfileString(SECTION, _T("Kanjidic"), _T("\\Program Files\\Kanjidic.euc"));
	CDict kanjidict;
	if (!initSearch(kanjidict.InitDict(dict, _T(""))))
		return;

	CString result;
	
	// lookup Dictionary
	if (m_Input.GetAt(0)>=0x4000) { // Kanji, lookup each one
		int c = 0;
		for (int i=0; i<m_Input.GetLength(); i++) {
			CString lookupstring = m_Input.Mid(i,1);
			if (lookupstring.GetAt(0)>=0x4000) { // kanji
				result = kanjidict.Find(3, lookupstring);
				while (c<600 && result.GetLength()!=0) {
					c++;
					m_Output += result;
					if (c<12 || c%20==0)
						OutputArea.SetWindowText(m_Output);
					result = kanjidict.Find(2, lookupstring);
				}
			}
		}
	}
	else { // Kana, English, or something else, Just lookup
		int c = 0;
		result = kanjidict.Find(3, m_Input);
		while (c<600 && result.GetLength()!=0) {
			c++;
			m_Output += result;
			if (c<12 || c%20==0)
				OutputArea.SetWindowText(m_Output);
			result = kanjidict.Find(2, m_Input);
		}
	}

	if (m_Output.GetLength()==0) {
		m_Output = _T("Nothing found.");
	}

	OutputArea.SetWindowText(m_Output);
	InputArea.SetEditSel(0,-1);
	InputArea.SetFocus();
}

void CChildView::OnAltdict() 
{
	CWaitCursor wait;
	// Get file path and name from the registry and then open the dictionaries
	CString dict, index;
    dict = theApp.GetProfileString(SECTION, _T("Alt"), _T("\\Program Files\\enamdict.euc"));
    index = theApp.GetProfileString(SECTION, _T("AltIndex"), _T("\\Program Files\\enamdict.jdx"));
	CDict namedict;
	if (!initSearch(namedict.InitDict(dict, index)))
		return;

	// Check type of input string
	inputType=2; // kana
	if (m_Input.GetAt(0)>=0x4000) { // Kanji
		inputType=0;
	}
	if (m_Input.GetAt(0)<=0x0100) { // Roman
		inputType=2;
	}

	// lookup Dictionary
	int c = 0;
	CString result = namedict.Find(inputType+1, m_Input);
	while (c<600 && result.GetLength()!=0) {
		c++;
		m_Output += result;
		if (c<12 || c%20==0)
			OutputArea.SetWindowText(m_Output);
		result = namedict.Find(inputType, m_Input);
	}

	if (m_Output.GetLength()==0) {
		m_Output = _T("Nothing found.");
	}

	OutputArea.SetWindowText(m_Output);
	InputArea.SetEditSel(0,-1);
	InputArea.SetFocus();
}

// try to lookup a verb or adjective from the modified form
// Does not handle irregular verbs

void CChildView::OnVerb() 
{
	struct {
		TCHAR verbend;    // last character of dictionary form of verb or adjective
		TCHAR * nextchar; // list of possible first characters of modified ending
	} verb[10] = {
// godan verbs
		/* 0 */	{ _T(''),_T("킢") },
		/* 1 */	{ _T(''),_T("") },
		/* 2 */	{ _T(''),_T("") },
		/* 3 */	{ _T(''),_T("") },
		/* 4 */	{ _T(''),_T("ĂƂ") },
		/* 5 */	{ _T(''),_T("΂тԂׂڂ") },
		/* 6 */	{ _T(''),_T("Ȃɂʂ˂̂") },
		/* 7 */	{ _T(''),_T("܂݂ނ߂") },
// ichidan or godan verb
		/* 8 */	{ _T(''),_T("ĂȂɂ͂܂") },
// pure adjective
		/* 9 */	{ _T(''),_T("") } 
	};
	
	CWaitCursor wait;
	// Get file path and name from the registry and then open the dictionaries
	CString dict, index;
    dict = theApp.GetProfileString(SECTION, _T("EDict"), _T("\\Program Files\\edict.euc"));
    index = theApp.GetProfileString(SECTION, _T("EDictIndex"), _T("\\Program Files\\edict.jdx"));
	CDict edict;
	if (!initSearch(edict.InitDict(dict, index)))
		return;

	WCHAR lookupstring[256];
	CString result;

	int inputlength = m_Input.GetLength();
	for (int i=inputlength-1; i>=1; i--) {
		for (int j=0; j<10; ++j) {
			wcscpy(lookupstring,m_Input.GetBuffer(0)); // wide string copy
			m_Input.ReleaseBuffer();
			WCHAR next = lookupstring[i];
			if (wcschr(verb[j].nextchar,next)==NULL) // wide char in string
				continue;
			lookupstring[i]=verb[j].verbend;
			lookupstring[i+1]=0;
			result = edict.Find(7, lookupstring);
			if (result.GetLength()!=0) {
				if (j < 8) {
					m_Output += _T("\r\nMaybe ");
					m_Output += CString(verb[j].verbend);
					m_Output += _T(" godan verb ");
				}
				else if (j == 8)
					m_Output += _T("\r\nMaybe ichidan or godan  verb ");
				else
					m_Output += _T("\r\nMaybe  adjective ");
				m_Output += _T("\r\n");
				while (result.GetLength()!=0) {
					m_Output += result;
					result = edict.Find(6, lookupstring);
				}
			}
		}
		if (m_Output.GetLength()!=0) {
			break;
		}
	}
	
	if (m_Output.GetLength()==0) {
		m_Output = _T("Nothing found.");
	}

	OutputArea.SetWindowText(m_Output);
	InputArea.SetEditSel(0,-1);
	InputArea.SetFocus();
}

// Normal cut, copy and paste commands
// For building up compound kanji

void CChildView::OnEditCut() 
{
	if (GetFocus() == (CWnd*)&OutputArea) {
		OutputArea.Copy();
	}
	else {
		InputArea.Cut();
	}
}

void CChildView::OnEditCopy() 
{
	if (GetFocus() == (CWnd*)&OutputArea) {
		OutputArea.Copy();
	}
	else {
		InputArea.Copy();
	}
}

void CChildView::OnEditPaste() 
{
	InputArea.Paste();
	InputArea.SetFocus();
}

// Handle resize of window

void CChildView::OnSize(UINT nType, int cx, int cy) 
{
	CWnd ::OnSize(nType, cx, cy);
	if (cx>50 && cy>50) { 
		InputArea.MoveWindow(CRect(0, 0, cx-0, 100));
		RECT iRect;
		InputArea.GetClientRect(&iRect);
		OutputArea.MoveWindow(CRect(0, iRect.bottom, cx-0, cy-0));	
	}
}

// Search the last results

void CChildView::OnSearchresults() 
{
	CWaitCursor wait;
	CDict tmpdict;
	if (!initSearch(tmpdict.MakeDict(m_Output)))
		return;

	// Check type of input string
	inputType=2; // kana
	if (m_Input.GetAt(0)>=0x4000) { // Kanji
		inputType=0;
	}
	if (m_Input.GetAt(0)<=0x0100) { // Roman
		inputType=2;
	}

	int c = 0;
	CString result = tmpdict.Find(inputType+1, m_Input);
	while (c<600 && result.GetLength()!=0) {
		c++;
		m_Output += result;
		if (c<12 || c%20==0)
			OutputArea.SetWindowText(m_Output);
		result = tmpdict.Find(inputType, m_Input);
	}

	if (m_Output.GetLength()==0) {
		m_Output = _T("Nothing found.");
	}

	OutputArea.SetWindowText(m_Output);
	InputArea.SetEditSel(0,-1);
	InputArea.SetFocus();
}

void CChildView::OnDictionaries() 
{
	CDictOptions optionDlg;
	optionDlg.DoModal();
	InputArea.SetEditSel(0,-1);
	InputArea.SetFocus();
}

