// Dict.cpp : implementation file
//
#include "stdafx.h"
#include "resource.h"
#include "Dict.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

extern WCHAR EUC2Unicode(WCHAR jiscode);
extern WCHAR Unicode2EUC(WCHAR c);

/////////////////////////////////////////////////////////////////////////////
// CDict

CDict::CDict()
{
	length = 0;
	dictionary = NULL;
	format = 0;
	memfile = false;
	indexLength = 0;
	index = NULL;
}

CDict::~CDict()
{
	if (memfile) {
		if (dictionary!=NULL)
			delete dictionary;	// release storage
		if (index!=NULL)
			delete index;		// release storage
	}
	else {
		if (dictionary!=NULL) {
			UnmapViewOfFile(dictionary); // free memory
			CloseHandle(hDict);
		}
		if (index!=NULL) {
			UnmapViewOfFile(index);		// free memory
			CloseHandle(hIndex);
		}
	}
}

// Make a temporary dictionary for searching 

bool CDict::MakeDict(CString file)
{
	memfile = true;
	length = file.GetLength()*2 + 2;
	dictionary = new unsigned char[length];
	unsigned char * p = dictionary;
	*p++ = 0xFF;
	*p++ = 0xFE;
	for (int i=0; i<file.GetLength(); ++i) {
		WCHAR c = file.GetAt(i);
		*p++ = c & 0x00FF;
		*p++ = (c>>8) & 0x00FF;
	}
	format = FormatUTF16;
	return true;
}

// Make a file into a dictionary

bool CDict::InitDict(CString file, CString indexfile)
{
	// First close the files if they are already open
	if (memfile) {
		if (dictionary!=NULL)
			delete dictionary;	// release storage
		if (index!=NULL)
			delete index;		// release storage
	}
	else {
		if (dictionary!=NULL) {
			UnmapViewOfFile(dictionary); // free memory
			CloseHandle(hDict);
		}
		if (index!=NULL) {
			UnmapViewOfFile(index);		// free memory
			CloseHandle(hIndex);
		}
	}
	// Initialize, same as create
	length = 0;
	dictionary = NULL;
	format = 0;
	memfile = false;
	indexLength = 0;
	index = NULL;

	// Next open the files 
	if (file.GetLength()==0)
		return false;
	CFile df;
	CFileException e;
#ifdef UNDER_CE
	memfile = false;
	if (!df.Open(file,CFile::modeRead,&e)) 
		return false;
	length = df.GetLength();	// get filesize of dictionary
	df.Close();
	HANDLE hmf = CreateFileForMapping(file, GENERIC_READ, FILE_SHARE_READ,
		NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	hDict = CreateFileMapping(hmf, NULL, PAGE_READONLY, 0, 0, NULL);
	dictionary = (unsigned char *)MapViewOfFile(hDict, FILE_MAP_READ, 0, 0, 0);
	CloseHandle(hmf);
	if (indexfile.GetLength()==0) {
		index = NULL;
	}
	else if (df.Open(indexfile,CFile::modeRead,&e)) {
		indexLength = df.GetLength() / 4;	// get filesize of index, 4 bytes per pointer
		df.Close();
		HANDLE hmf = CreateFileForMapping(indexfile, GENERIC_READ, FILE_SHARE_READ,
			NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
		hIndex = CreateFileMapping(hmf, NULL, PAGE_READONLY, 0, 0, NULL);
		index = (long *)MapViewOfFile(hIndex, FILE_MAP_READ, 0, 0, 0);
		CloseHandle(hmf);
	}
#else
	memfile = true;
	if (!df.Open(file,CFile::modeRead,&e))
		return false;
	length = df.GetLength();			// get filesize of dictionary
	dictionary = new unsigned char[length];
	df.Read(dictionary,length);			// read file
	df.Close();
	if (indexfile.GetLength()) {
		index = NULL;
	}
	else if (df.Open(indexfile,CFile::modeRead,&e)) {
		indexLength = df.GetLength() / 4;	// get filesize of index, 4 bytes per pointer
		index = new long[indexLength];
		df.Read(index,indexLength);			// read file
		df.Close();
	}
#endif
	// Find format of dictionary file
	if (*dictionary == 0xFF && *(dictionary+1)==0xFE) // UTF16LE
		format = FormatUTF16;
	else if (*dictionary == 0xEF && *(dictionary+1)==0xBB && *(dictionary+2)==0xBF) // UTF8
		format = FormatUTF8;
	else if (*dictionary == 0xA1 && *(dictionary+1)==0xA9) // EUC Edict and Enamedic
		format = FormatEUC;
	else if (*dictionary == 0x23 && *(dictionary+1)==0x20) // EUC Kanjidic
		format = FormatEUC;
	else // Default is EUC
		format = FormatEUC;

	if (length==0 || dictionary==NULL)
		return false;
	return true;
}

// find str within the dictionary file
// Find returns one line at a time
// searchflags = 1 to start a new search, 0 to continue a search
// searchflags = 2 to match start of word
// searchflags = 4 to match end of word

CString CDict::Find(int searchflags, CString str)
{
	if (length==0 || dictionary==NULL)
		return _T("No dictionary found.");

	WCHAR buffer[256];
	int ln = str.GetLength();
	if (ln > 250)
		ln = 250;

	// Convert to hiragana and lowercase
	WCHAR * pstr = buffer;
	for (int i=0; i<ln; i++) {
		WCHAR c = str.GetAt(i);
		if (c>=0x30A0 && c<=0x30FF) // Katakana to Hiragana
			c -= 0x60;
		if ((c >= 'A') && (c <= 'Z')) // capital to lowercase 
			c |= 0x20;
		*pstr++ = c;
	}
	*pstr++ = 0;

	if (format == FormatEUC) {
		// EUC format dictionary, convert input to EUC code
		for (int i=0; i<ln; i++) {
			buffer[i] = Unicode2EUC(buffer[i]);
		}
	}

	if (indexLength==0 || index==NULL) 
		return FindNoIndex(searchflags, buffer);
	else
		return FindWithIndex(searchflags, buffer);
}

/////////////////////////////////////////////////////////////////////////////
// EUC Find first occurance of the 2nd string within the first string,
// ignoring case and matching katakana and hirgana
// Only used by FindNoIndex
// The first string is the dictionary, in EUC, UTF8 or UTF16LE
// The second string is the user entry in EUC16 or UTF16

unsigned char * CDict::Kstrstr(unsigned char * p, WCHAR * q)
{
	if (format==FormatEUC)
		return KstrstrEUC(p, q);
	else if (format==FormatUTF8)
		return KstrstrUTF8(p, q);
	else // UTF16)
		return (unsigned char *)KstrstrUTF16((WCHAR *)p, q);
}

unsigned char * CDict::KstrstrEUC(unsigned char * p, WCHAR * q)
{
	WCHAR c2;
	WCHAR c1;

	while (*p) {
		int i=0;
		int j=0;
		while(true) {
			c1 = p[i++];
			c2 = q[j++];
			if (c2==0)
				return p;
			if (c1>0x7F) { // Double byte, high byte first
				if (p[i]==0)
					c1 = 0;
				else
				    c1 = (c1<<8) | p[i++];
			}
			if ((c1&0xFF00)==0xA500) // Katakana to Hiragana
				c1 &= 0xA4FF;
			if ((c1 >= 'A') && (c1 <= 'Z')) // capital to lowercase
				c1 |= 0x20;
			if (c1 != c2 ) 
				break;
		}
		if (*p>0x7F) // double byte character
			p+=2;
		else		// single byte character
			p++;
	}
	return NULL;
}

unsigned char * CDict::KstrstrUTF8(unsigned char * p, WCHAR * q)
{
	WCHAR c1, c2;
	while (*p) {
		int i=0; 
		int j=0;
		while(true) {
			c1 = p[i++];
			c2 = q[j++];
			if (c2==0)
				return p;
			if (c1>=0xE0) { // Triple byte, high byte first
				c1 = ((c1&0x0F)<<12) | ((p[i++]&0x3F)<<6);
				c1 |= (p[i++]&0x3F);
			}
			else if (c1>=0x80) { // Double byte, high byte first
				c1 = ((c1&0x1F)<<6) | (p[i++]&0x3F);
			}
			else if (c1>=0x80) { // illegal byte
				c1 = 0;
			}

			if (c1>=0x30A0 && c1<=0x30FF) // Katakana to Hiragana
				c1 -= 0x60;
			if ((c1 >= 'A') && (c1 <= 'Z')) // capital to lowercase 
				c1 |= 0x20;

			if (c1 != c2 ) 
				break;
		}
		p++;
	}
	return NULL;
}

WCHAR * CDict::KstrstrUTF16(WCHAR * p, WCHAR * q)
{
	WCHAR c1, c2;

	while (*p) {
		for (int i = 0; ; i++) {
			c1 = p[i];
			c2 = q[i];
			if (c2==0)
				return p;
			if (c1>=0x30A0 && c1<=0x30FF) // Katakana to Hiragana
				c1 -= 0x60;
			if ((c1 >= 'A') && (c1 <= 'Z')) // capital to lowercase 
				c1 |= 0x20;
			if (c1 != c2 ) 
				break;
		}
		p++;
	}
	return NULL;
}

/////////////////////////////////////////////////////////////////////////////
// Special strcmp, ignore case etc, Only used by FindWithIndex
// First parameter is the dictionary
// Second is the search string

int CDict::Kstrcmp(unsigned char * p, WCHAR * q)
{
	if (format==FormatEUC) {
		return KstrcmpEUC(p, q);
	}
	else if (format==FormatUTF8) {
		return KstrcmpUTF8(p, q);
	}
	else { // UTF16
		return KstrcmpUTF16((WCHAR *)p, q);
	}
}

int CDict::KstrcmpEUC(unsigned char * p, WCHAR * q)
{
	WCHAR c1,c2;
	while(true) {
		c1 = *p++;
		c2 = *q++;
		if (c1==0 || c2==0 ) // end of string marker
			return 0;
		if (c1>0x7F) { // Double byte, high byte first
			c1 = (c1<<8) | *p++;
		}

		if ((c1&0xFF00)==0xA500) // Katakana to Hiragana
			c1 &= 0xA4FF;
		if ((c1 >= 'A') && (c1 <= 'Z')) // capital to lowercase
			c1 |= 0x20;
		if (c1 != c2 ) 
			break;
	}
	return c1-c2;
}

int CDict::KstrcmpUTF8(unsigned char * p, WCHAR * q)
{
	WCHAR c1,c2;
	while(true) {
		c1 = *p++;
		c2 = *q++;
		if (c1==0 || c2==0 ) // end of string marker
			return 0;

		if (c1>=0xE0) { // Triple byte, high byte first
			c1 = ((c1&0x0F)<<12) | (((*p++)&0x3F)<<6);
			c1 |= ((*p++)&0x3F);
		}
		else if (c1>=0x80) { // Double byte, high byte first
			c1 = ((c1&0x1F)<<6) | ((*p++)&0x3F);
		}
		else if (c1>=0x80) { // illegal byte
			c1 = 0;
		}

		if ((c1&0xFF00)==0xA500) // Katakana to Hiragana
			c1 &= 0xA4FF;
		if ((c1 >= 'A') && (c1 <= 'Z')) // capital to lowercase
			c1 |= 0x20;
		if (c1 != c2 ) 
			break;
	}
	return c1-c2;
}

int CDict::KstrcmpUTF16(WCHAR * p, WCHAR * q)
{
	WCHAR c1,c2;
	for (int i = 0; ; i++) {
		c1 = p[i];
		c2 = q[i];
		if (c1==0 || c2==0 )
			return 0;
		if (c1>=0x30A0 && c1<=0x30FF) // Katakana to Hiragana
			c1 -= 0x60;
		if ((c1 >= 'A') && (c1 <= 'Z')) // capital to lowercase 
			c1 |= 0x20;
		if (c1 != c2 ) 
			break;
	}
	return c1-c2;
}

/////////////////////////////////////////////////////////////////////////////
// Output a line from the dictionary
// p1 points top some point in the line
// Returns the complete line

CString CDict::OutputLine(unsigned char * p1)
{
	unsigned char * dict = dictionary;
	if (format==FormatUTF16) // Skip 2 bytes BOM
		dict += 2;
	else if (format==FormatUTF8) // Skip 3 bytes BOM
		dict += 3;

	// move p1 back to the start of the line, after newline character
	while(p1>dict) {
		if (format==FormatUTF16) { // UTF16 must be byte pair
			if (*((WCHAR *)p1) == '\n') {
				p1 += 2;
				break;
			}
			p1 -= 2;
		}
		else { // EUC or UTF8 is byte at a time
			if (*p1 == '\n') {
				p1 += 1;
				break;
			}
			p1 -= 1;
		}
	}

	unsigned char * p = p1;
	unsigned char * dictend = dictionary + length;

	// move p forward to the end of line, point to return or newline
	while(p<dictend) {
		if (format==FormatUTF16) { // UTF16 must be byte pair
			if (*((WCHAR *)p)=='\n' || *((WCHAR *)p)=='\r')
				break;
			p += 2;
		}
		else { // EUC or UTF8 is byte at a time
			if (*p=='\n' || *p=='\r')
				break;
			p += 1;
		}
	}

	WCHAR c;
	CString doc = "";

	// copy chars in range (p1 to p) to the output string 
	while(p1<p) {
		if (format==FormatEUC) { // EUC 1 or 2 bytes
			c = *p1++;
			if (c>=0x80)
				c = (c<<8) | *p1++;
			c = EUC2Unicode(c);
		}
		else if (format==FormatUTF8) { // UTF8 1 to 3 bytes
			c = *p1++;
			if (c>=0xE0) { // Triple byte, high byte first
				c = ((c&0x0F)<<12) | (((*p1++)&0x3F)<<6);
				c |= ((*p1++)&0x3F); // must be after because order of eval ++ is not fixed
			}
			else if (c>=0xC0) { // Double byte, high byte first
				c = ((c&0x1F)<<6) | ((*p1++)&0x3F);
			}
			else if (c>=0x80) { // illegal byte
				c = '?';
			}
		}
		else { // UTF16 always 2 bytes
			c = *((WCHAR *)p1);
			p1 += 2;
		}
		doc += (WCHAR)c;
	}
	// delete leading spaces
	while (doc.GetAt(0)==' ') {
		doc = doc.Right(doc.GetLength()-1);
	}
	return _T("  ") + doc + _T("\r\n");
}

/////////////////////////////////////////////////////////////////////////////
// Calc number of bytes for the search string, but in the dictionary format

int CDict::WordLength(WCHAR *str)
{
	if (format==FormatEUC) {
		// Dictionary EUC, search string in EUC
		int i = 0;
		while (*str) {
			if (*str <= 0x007F)
				i += 1;
			else 
				i += 2;
			str++;
		}
		return i;
	}
	else if (format==FormatUTF8) {
		// Dictionary UTF8, search string in UTF16
		int i = 0;
		while (*str) {
			if (*str <= 0x007F)
				i += 1;
			else if (*str <= 0x07FF)
				i += 2;
			else 
				i += 3;
			str++;
		}
		return i;
	}
	else { // UTF16
		// Dictionary UTF16, search string in UTF16
		int i = 0;
		while (*str++) {
			i += 2;
		}
		return i;
	}
}

/////////////////////////////////////////////////////////////////////////////
// Find without an index file

CString CDict::FindNoIndex(int searchflags, WCHAR * str)
{
	static unsigned char * p;
	if ((searchflags&1)==1) {
		if (format==FormatEUC)
			p = dictionary;
		else if (format==FormatUTF8)
			p = dictionary + 3;	// Ignore BOM bytes
		else if (format==FormatUTF16)
			p = dictionary + 2;	// Ignore BOM bytes
	}
	unsigned char * dictend = dictionary + length;
	unsigned char * p1 = p;

	while(p < dictend) {
		// find next word
		p1 = Kstrstr(p,str); 
		if (p1==NULL) { // nothing found
			return _T("");
		}
		p = p1 + WordLength(str); // move p to end of word

		WCHAR c2;
		if (format==FormatUTF16) 
			c2 = *(WCHAR *)(p1-2); // character before start of word
		else // UTF8 or EUC
			c2 = *(p1-1); // character before start of word

		// check start of word ?
		if ((searchflags&2)==0 || c2==' ' || c2=='[' || c2=='{' || c2=='/' || c2=='\n') {
			WCHAR c;
			if (format==FormatUTF16) {
				c = *(WCHAR *)p; // character after end of word
			}
			else { // EUC or UTF8
				c = *p; // character after end of word
			}
			// check end of word ?
			if ((searchflags&4)==0 || c==' ' || c==']' || c=='}' || c=='/' || c=='\n') {
				// move p to end of line, point to newline
				while(p<dictend) {
					if (format==FormatUTF16) {
						if (*(WCHAR *)p == '\n')
							break;
						p += 2;
					}
					else { // EUC or UTF8
						if (*p == '\n')
							break;
						p += 1;
					}
				}

				return OutputLine(p1);
			}
		}
	}
	return _T("");
}

/////////////////////////////////////////////////////////////////////////////
// Find using an index file

CString CDict::FindWithIndex(int searchflags, WCHAR * str)
{
	long lo, hi;
	int res;
	static long it;

	if ((searchflags&1)==1)	{ // binary search of index table
		hi = indexLength;
		lo = 1;
		for (int iterlimit=30; iterlimit>0; --iterlimit) {
  			it = (lo+hi)/2;
			res = Kstrcmp(&dictionary[index[it]-1], str);
			if (res == 0)
				break;
			if (res > 0)
				hi = it-1;
			else
				lo = it+1;
			if (lo > hi) 
				break;
		} // end loop
		if (res != 0) { // not found
			return _T("");
		}

		// step back to the first entry
		while (it >= 2) {
			res = Kstrcmp(&dictionary[index[it-1]-1], str);
			if (res == 0)
				it--;
			else
				break;
		}
	}
	
	// 'it' points to next index entry
	unsigned char * p1 = &dictionary[index[it]-1];
	it++;
	if (Kstrcmp(p1, str)!=0) {
		return "";
	}
	if ((searchflags&4)==4) { // check whole word match, ie next character is not part of the word
		WCHAR c;
		if (format==FormatUTF16) {
			c = *(WCHAR *)(p1 + WordLength(str)); // character after end of word
		}
		else { // EUC or UTF8
			c = *(p1 + WordLength(str)); // character after end of word
		}
		if (c!=' ' && c!=']' && c!='}' && c!='/' && c!='\n')
			return _T(""); // Since the list is sorted there are no more entries
	}
	return OutputLine(p1);
}
 
/////////////////////////////////////////////////////////////////////////////
// CDict message handlers
