// KanjiPen.cpp: implementation of the CKanjiPen class.
//
//////////////////////////////////////////////////////////////////////

#include <windows.h>
#include "KanjiPen.h"
 
#pragma intrinsic( labs , abs )

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

#define GRIDSIZE 70

#define diAngCostScale      ((long)100)	// See angles.pl for derivation.
#define diHugeCost          ((long)32*diAngCostScale)

#define diMaxScoreToSquare ((long)0x7fff)
#define diMaxScoreSquared  ((long)(diMaxScoreToSquare*diMaxScoreToSquare))

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CKanjiPen::CKanjiPen()
{
	StrokeData = NULL;
	const TCHAR * file = _T("\\Program Files\\Strokedata.txt");
	HANDLE hmf = CreateFileForMapping(file, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	HANDLE hfm = CreateFileMapping(hmf, NULL, PAGE_READONLY, 0, 0, NULL);
	StrokeData = (TCHAR *)MapViewOfFile(hfm, FILE_MAP_READ, 0, 0, 0);
}

CKanjiPen::~CKanjiPen()
{
	if (StrokeData!=NULL)
		UnmapViewOfFile(StrokeData);		
}

int Angle32(int xdif, int ydif) 
{
	bool xneg, yneg, xyflip;
	int i32nd, xtmp, islope;

    if ((xneg = (xdif < 0)))
		xdif = -xdif;
	if ((yneg = (ydif < 0)))
		ydif = -ydif;
	if ((xyflip = (ydif < xdif))) {
		xtmp = xdif; xdif = ydif; ydif = xtmp;
	}

    if (xdif == 0) {
		if (ydif == 0)
			return 32;
		else
			i32nd = 0;
	}
    else {
		/* The 4 comparison values were generated with the accompanying
		 * perl script, then open coded here for speed and reasonable
		 * space efficiency.  The values were chosen to make the results 
		 * match those of atan2 in rounded double
		 * precision floating point.  -rwells, 970713.
		 */

		islope = (100 * xdif) / ydif;
		if (islope < 54) {             /* test #2, first test. */
			if (islope < 10)           /* test #0, second test. */
				i32nd = 0;             /*          got  #0 after 2 tests. */
			else if (islope < 31)      /* test #1, third test. */
				i32nd = 1;             /*          got  #1 after 3 tests. */
			else
				i32nd = 2;             /*          got  #2 after 3 tests. */
		}
		else if (islope < 83)          /* test #3, second test. */
			i32nd = 3;                 /*          got  #3 after 2 tests. */
		else
			i32nd = 4;                 /*          got  #4 after 2 tests. */
	}    

	if (xyflip)
		i32nd = (8 - i32nd);
	if (yneg)
		i32nd = (16 - i32nd);
	if (xneg)
		i32nd = (32 - i32nd);

    return i32nd % 32;
}

long CKanjiPen::StrokeScorerEvalItem(TCHAR * cpEntry, CStroke m_strokes[40], int m_nStrokeNo, long lBestScore) {
	TCHAR * cp = cpEntry;
	int    iStroke;
	TCHAR cpPath[40];
	TCHAR cpDrawn[40];
	TCHAR * cpPathEnd;
	long   iThisScore;
	long   iScore = 0;
	const TCHAR dir[9] = {
	//    A   B    C    D   E   F   G    H    I
	 	 20, 16,  12,  24,  8,  8, 28,   0,   4 
	//  7:30 6:00 4:30 9:00  3:00  10:30 12:00 1:30
	};

	/* Loop through stroke descriptions */
	for (iStroke = 0; iStroke<=m_nStrokeNo; iStroke++) {

		cpPathEnd = cpPath;
		if (*cp>='A' && *cp<='I')
			*cpPathEnd++ = dir[*cp-'A'];
		else 
			break; // NoMoreStrokes

		for (cp++; ; cp++) {	/* Loop through following chars for stroke */
			if (*cp>='a' && *cp<='i')
				*cpPathEnd++ = dir[*cp-'a'];
			else 
				break; // ThisStrokeDone
		} /* end loop through chars for stroke */

		// convert to angle32
		for (int i=1; i<m_strokes[iStroke].iLen; i++) {
			int x = m_strokes[iStroke].m_point[i].x - m_strokes[iStroke].m_point[i-1].x;
			int y = m_strokes[iStroke].m_point[i-1].y - m_strokes[iStroke].m_point[i].y;
			cpDrawn[i-1] = Angle32(x,y);
		}

		iThisScore = ScoreStroke(cpDrawn, m_strokes[iStroke].iLen-1, cpPath, (cpPathEnd - cpPath), 0 /*depth*/);
		//iThisScore = StrokeDicScoreStroke(m_strokes[iStroke].m_point, m_strokes[iStroke].iLen, cpPath, (cpPathEnd - cpPath), 0 /*depth*/);
		
		if (iThisScore < 0)
			iThisScore = 0;
		if (iThisScore >= diMaxScoreToSquare)
			iThisScore = diMaxScoreSquared + 1;
		else
			iThisScore = (iThisScore * iThisScore);

		iScore += iThisScore;
		if (iScore < 0 )
			iScore = diMaxScoreSquared + 2;

		if (iScore > lBestScore) {
			//TRACE(_T("."));
			return iScore;
		}
	} /* end loop through stroke descriptions */
//NoMoreStrokes:
	// now iscore * iscore is returned

    /* Handle optional extra filters... may modify *ipScore, updates cp. */
	if (*cp == '|') {
		long r = StrokeScorerExtraFilters(m_strokes, cp+1);
		if (r>0) {
			iScore += r * r;
			if (iScore < 0)
				iScore = diMaxScoreSquared + 3;
		}
		else {
			iScore -= r * r;
			if (iScore < 0)
				iScore = 0;
		}
	}
	return iScore;
}

/* ----- StrokeDicScoreStroke ---------------------------------------------- */

long CKanjiPen::StrokeDicScoreStroke(POINT * points, int iLen, TCHAR * cpPath, int iPathLen, int iDepth) {
	long iScore, iThisScore;
	int iMid, iStep, iPathMid, iPathRest;
	long iDifX, iDifY;
	int iAng32, iPath32, iDif32;

	if (iLen < 2 || iPathLen < 1)
		return diHugeCost+1;

	if (iPathLen == 1) { // Single-line segment candidate stroke
		iDifX = points[iLen-1].x - points[0].x;
		iDifY = points[0].y - points[iLen-1].y; /* Flip from display to math axes. */

		if (iDifX == 0 && iDifY == 0) /* Two samples at same place... */
			return diHugeCost+2;

		// Subdivide recursively while stroke is long and depth is shallow.
		if (iLen > 2 && iDepth < 5) {
			iMid = iLen >> 1;

			/* Note that we use the middle point on both sides... */
			iScore  = StrokeDicScoreStroke(points, iMid+1, cpPath, iPathLen, iDepth+1) >> 1;
			iScore += StrokeDicScoreStroke(&points[iMid], iLen-iMid, cpPath, iPathLen, iDepth+1) >> 1;
			return iScore;

		} /* End if stroke is long, and depth is shallow... */

		/* Time to score this segment against desired direction. */
		// Assume a straight line, even though its not		
		iAng32 = Angle32(iDifX, iDifY);
		iPath32 = *cpPath;

		if (iAng32 >= iPath32)
			iDif32 = iAng32 - iPath32;
		else
			iDif32 = iPath32 - iAng32;

		return iDif32 * diAngCostScale;

	} /* end if path len is 1. */
	else { // Multi-line segment candidate stroke
		iScore = diHugeCost * 10;
		// Divide candidate stroke in half
		iPathMid = iPathLen >> 1;
		iPathRest = iPathLen - iPathMid;
		
		// divide drawn stroke into small pieces
		if (iLen < 10)
			iStep = 1;
		else
			iStep = iLen >> 3;
		int iLenMid = iPathMid;
		int iLenMax = iLen - iPathRest;
		if (iLenMid >= iLenMax) { // not enough drawn segments
			iLenMid = 1;
			iLenMax = iLen;
		}
		for (iMid = iLenMid; iMid < iLenMax; iMid += iStep) {
			iThisScore  = StrokeDicScoreStroke(points, iMid+1, cpPath, iPathMid, iDepth+1) >> 1;
			iThisScore += StrokeDicScoreStroke(&points[iMid], iLen-iMid, cpPath+iPathMid, iPathRest, iDepth+1) >> 1;
			if (iThisScore < iScore)
				iScore = iThisScore;
		} /* end for trials on various mid-point divisions of stroke... */

		return iScore;
	} /* end if path len is >1. */
}

/* ----- ScoreStroke ---------------------------------------------- */

long CKanjiPen::ScoreStroke(TCHAR * p1, int iL1, TCHAR * p2, int iL2, int iDepth) {
	long iScore, iThisScore;
	int iMid, iPathMid, iPathRest;
	int iAng32, iPath32, iDif32;

	// shortest in iL1, p1
	if (iL2 < iL1) {
		TCHAR * t = p1; p1 = p2; p2 = t;
		int l = iL1; iL1 = iL2; iL2 = l;
	}

	if (iL1 < 1)
		return diHugeCost+1;

	if (iL1 == 1) { // Single-segment stroke matching 1 or more segments
		iThisScore = 0;
		iPath32 = *p1;
		for (int i=0; i<iL2; i++) {
			iAng32 = p2[i];
			if (iAng32 >= iPath32)
				iDif32 = iAng32 - iPath32;
			else
				iDif32 = iPath32 - iAng32;
			iThisScore += iDif32;
		}
		return iThisScore * diAngCostScale / iL2;
	}
	else { // Multi-segment stroke
		iScore = diHugeCost * 10;
		// Divide long stroke in half
		iPathMid = iL2 >> 1;
		iPathRest = iL2 - iPathMid;
		
		// divide short stroke into small pieces
		for (iMid = 1; iMid < iL1; iMid++) {
			iThisScore  = ScoreStroke(p1, iMid, p2, iPathMid, iDepth+1) >> 1;
			iThisScore += ScoreStroke(&p1[iMid], iL1-iMid, &p2[iPathMid], iPathRest, iDepth+1) >> 1;
			if (iThisScore < iScore)
				iScore = iThisScore;
		} /* end for trials on various mid-points */
		return iScore;
	} /* end if path len is >1. */
}

/* ----- StrokeScorerExtraFilters ---------------------------------------------*/

// Extra Filter x1-x2  means start of 1 is to the left of 2
//              i            end of
//              a            ave of
//              y1-y2        start of 1 is below 2
//              j            end of
//              b            ave of
//              l1-l2        length of 1 is longer than 2 

//              y1<j2        start of 1 to end of 2 is < 1/2 GridSize

long CKanjiPen::StrokeScorerExtraFilters(CStroke * pStroke, TCHAR * cp) {
	TCHAR c;
	TCHAR cArg[2];
	int iStroke[2];
	long iDiff, iVal[2];
	int idx = 0;
	bool bMust = false;
	long iScore = 0;
	TCHAR fn;
	
	cArg[0] = cArg[1] = 0;
	iStroke[0] = iStroke[1] = 0;

    /* Simple parser for Filter strings. assumes a1-b1 structure,
	 * where a and b can be any single alphabetic cmd char, the 
	 * numbers can be multiple digit, and b1 can optionally be 
	 * followed by '!' to insist on the filter passing.  There can
	 * be multiple filters but they have to be separated by '!' or
	 * space(s).  The filter string is terminated by a null byte
	 * or an 8-bit char, the beginning of the next entry.  
	 * Leading spaces and trailing spaces are ignored. -rwells, 970722.
	 */

	for (c = *cp; true; cp++, c = *cp) {
		switch (c) {

		case 'x': 
		case 'y':
		case 'i':
		case 'j':
		case 'a':
		case 'b':
		case 'l':
			cArg[idx] = c;
			break;

		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9':
			iStroke[idx] = iStroke[idx] * 10 + (c - '0');
			break;

		case '-':	/* Switch to second arg when we see function sign. */
		case '<':
		case '=':
			fn = c;
			idx = 1;
			break;

		case '!':
			bMust = true;

		case ' ':
		default:
			/* If we are in the second argument, execute and reset. */
			if (idx == 1) {
				iVal[0] = StrokeScorerExtraEval(&pStroke[iStroke[0]-1], cArg[0]);
				iVal[1] = StrokeScorerExtraEval(&pStroke[iStroke[1]-1], cArg[1]);
				if (fn=='-') { // right, below, longer
					iDiff = (iVal[0] - iVal[1]);
					if (iDiff < 0 && bMust) {
						return 9999999;
					}
					iScore -= iDiff;
				}
				else if (fn=='=') { // should be same
					iDiff = iVal[0] - iVal[1];
					iScore += iDiff * iDiff;
				}
				else if (fn=='<') { // small character
					iDiff = labs(iVal[1] - iVal[0]);
					if (iDiff < (long)(GRIDSIZE/2))
						iScore -= (long)GRIDSIZE;
					else 
						return 9999999;
				}

				/* Reset state for next filter... */
				idx = 0;
				bMust = false;
				cArg[0] = cArg[1] = 0;
				iStroke[0] = iStroke[1] = 0;
			}

			/* If this is a terminating char, break out of loop and return. */
			if (c == '\r' || c == '\0')
				goto ReturnNow;
		} /* end switch on char */
	} /* end for each char in filter spec... */
ReturnNow:
	return iScore;
}

/* ----- StrokeScorerExtraEval ------------------------------------------------*/

int CKanjiPen::StrokeScorerExtraEval(CStroke * pStroke, TCHAR cArg) {

	/* The coordinates are as we got them, screen
	 * coordinates with 0,0 in upper left, y increasing downward,
	 * x increasing to the right.  And the extra filters agree with this.
	 */

	switch(cArg) {
	case 'x':
		return pStroke->m_point[0].x;
	case 'y':
		return pStroke->m_point[0].y;
	case 'i':
		return pStroke->m_point[pStroke->iLen-1].x;
	case 'j':
		return pStroke->m_point[pStroke->iLen-1].y;
	case 'a':
		return ((pStroke->m_point[0].x + pStroke->m_point[pStroke->iLen-1].x) >> 1);
	case 'b':
		return ((pStroke->m_point[0].y + pStroke->m_point[pStroke->iLen-1].y) >> 1);
	case 'l':
		{  // very rough length of stroke
			int iVal = pStroke->m_point[0].x - pStroke->m_point[pStroke->iLen-1].x;
			int iVaiL2 = pStroke->m_point[0].y - pStroke->m_point[pStroke->iLen-1].y;
			return abs(iVal) + abs(iVaiL2);
		}
	}
	// TRACE(_T("Illegal character in extra filter"));
	return 0;					/* Unrecognized cArg cmd char. */
}
/* ----- end of scoring.c --------------------------------------------------*/

